# -*- coding: utf-8 -*-

'''
views
-----

defines field views for spacy objects.
'''


_DEP_SPAN = {
    'MR': (
        'vector_norm',
        'root.orth',
        'root.lemma',
        'root.dep',
        'root.head.orth',
        'root.head.lemma',
        'root.head.pos',
    ),
    'HR': (
        'text',
        'root.orth_',
        'root.lemma_',
        'root.dep_',
        'root.head.orth_',
        'root.head.lemma_',
        'root.head.pos_',
    ),
}

_DEP_TOKEN = {
    'MR': (
        'orth',
        'lemma',
        'dep',
        'head.orth',
        'head.lemma',
        'head.pos',
    ),
    'HR': (
        'orth_',
        'lemma_',
        'dep_',
        'head.orth_',
        'head.lemma_',
        'head.pos_',
    ),
}

_ENT_SPAN = {
    'MR': (
        'vector_norm',
        'label',
    ),
    'HR': (
        'text',
        'label_',
    ),
}

_POS_TOKEN = {
    'MR': (
        'orth',
        'pos',
        'lemma',
    ),
    'HR': (
        'orth_',
        'pos_',
        'lemma_',
    ),
}



# -*- coding: utf-8 -*-

'''
interface
---------

collection for building natural language app interfaces
'''


from functools import wraps
from inspect import getmembers

from spacy.en import English

from pynlai import core


class Trigger(object):
    '''
    natural language trigger where:
      call is a nl fcn from core that returns a list of spacy objects
      view is a list of keys to be used with core.create_view
      criteria is a list of (key, value) tuples to match in view
      callback is automatically set to decorated function by run
    '''

    def __init__(self, call, view, criteria, callback=None):
        self.call = call
        self.view = view
        self.criteria = criteria
        self.callback = callback


class Argument(Trigger):
    '''
    natural language trigger that is an argument to a decorated function
    where callback is a fcn to process view results and return a dict
    '''

    argument = True

    def __init__(self, call, view, criteria, callback):
        super(Argument, self).__init__(call, view, criteria, callback)


def nl_function(*triggers):
    '''
    decorator for registering a function in the pynlai
    '''

    def decorator(fcn):

        # inject triggers into decorated function
        setattr(fcn, '__pynlai_triggers', triggers)

        @wraps(fcn)
        def wrapper(*args, **kwargs):

            return fcn(*args, **kwargs)

        return wrapper

    return decorator


@core.nlp_preprocess(English)
def run(doc, nlp, obj):
    '''
    runs a natural language document through each registered object
    '''

    args = dict()
    callbacks = list()
    def command(*args, **kwargs): return None  # noqa

    # store callback for triggers meeting criteria
    for _, fcn in getmembers(obj):

        for t in getattr(fcn, '__pynlai_triggers', ()):

            for s in doc.sents:

                for e in t.call(doc=s.text, nlp=nlp):

                    nl_view = core.create_view(e, t.view)

                    if set(t.criteria.items()) <= set(nl_view.items()):
                        callback = t.callback or fcn
                        callbacks.append((t, s.text, callback))

    # handle stored callbacks
    for trigger, sentence, callback in callbacks:

        if hasattr(trigger, 'argument'):
            args.update(callback(sentence))

        else:
            command = callback  # noqa

    return command(**args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
test
----

unit tests for pynlai package
'''


import unittest
from click.testing import CliRunner

from pynlai import core
from pynlai import cli


class Test(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    def tearDown(self):
        pass

    def test_cli(self):
        result = self.runner.invoke(
            cli.main,
            catch_exceptions=False,
        )
        self.assertEqual(result.exit_code, 0)

    def test_cli_help(self):
        result = self.runner.invoke(
            cli.main,
            ['--help'],
            catch_exceptions=False,
        )
        self.assertEqual(result.exit_code, 0)
        self.assertTrue('Show this message and exit' in result.output)

    def test_cli_parse(self):
        pass



# -*- coding: utf-8 -*-

'''
cli
---


console script for pynlai.
'''


import pprint
from six.moves import input

import click
import en_core_web_sm as en

from pynlai.core import (
    create_view,
    to_ent,
    to_nc,
    to_obj,
    to_pos,
    to_sub,
)

from pynlai.views import (
    _DEP_SPAN,
    _DEP_TOKEN,
    _ENT_SPAN,
    _POS_TOKEN,
)


views = {
    'dep_span': _DEP_SPAN,
    'dep_token': _DEP_TOKEN,
    'ent_span': _ENT_SPAN,
    'pos_token': _POS_TOKEN,
}

func_view = {
    'ent': (to_ent, _ENT_SPAN['HR']),
    'nc': (to_nc, _DEP_SPAN['HR']),
    'obj': (to_obj, _DEP_TOKEN['HR']),
    'pos': (to_pos, _POS_TOKEN['HR']),
    'sub': (to_sub, _DEP_TOKEN['HR']),
}

pp = pprint.PrettyPrinter(indent=4, width=200)


@click.group()
@click.pass_context
@click.option(
    '--interactive', '-i',
    is_flag=True,
    help='interactive prompt for sentence input'
)
@click.option(
    '--pipeline', '-p',
    default=('pos',),
    multiple=True,
    type=click.Choice(func_view.keys()),
    help='nl function(s) and paired view to use in processing pipeline'
)
@click.option(
    '--view', '-v',
    multiple=True,
    type=click.Choice(views.keys()),
    help='override paired view(s) with view(s)'
)
def main(ctx, interactive, pipeline, view):
    '''
    pynlai command line interface
    '''

    if view:
        vs = [e for v in view for e in views[v]['HR']]
        fv = {k: (v[0], vs) for k, v in func_view.items()}
    else:
        fv = func_view

    ctx.obj['i'] = interactive
    ctx.obj['pipeline'] = [fv[k] for k in pipeline]
    ctx.obj['nlp'] = en.load()


@main.command()
@click.pass_context
@click.argument('sent', required=False)
def parse(ctx, sent):
    '''
    parse a sentence
    '''

    nlp = ctx.obj['nlp']

    while True:

        sent = input('parse>  ') if ctx.obj['i'] else sent

        try:

            for f, v in ctx.obj['pipeline']:
                r = [create_view(e, v) for e in f(doc=sent, nlp=nlp)]
                click.echo(pp.pprint((f.__name__, r)))

            if not ctx.obj['i']:
                break

        except KeyboardInterrupt:
            break


@main.command()
@click.pass_context
@click.argument('files', nargs=-1)
def parsefile(ctx, files):
    '''
    parse document(s), globs ok
    '''

    nlp = ctx.obj['nlp']

    for f in files:

        click.echo(f)

        with open(f, 'r') as f_obj:

            doc = nlp(f_obj.read())

            for sent in doc.sents:
                click.echo(sent)
                ctx.invoke(parse, sent=sent.text)


def entry_point():
    '''
    required to make setuptools and click play nicely (context object)
    '''

    return main(obj={})


if __name__ == "__main__":
    entry_point()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
test_interface
--------------

unit tests for pynlai interface module
'''


from collections import OrderedDict
import six
import unittest

import en_core_web_sm as en

import pynlai
from pynlai import core
from pynlai import views

from .shared import *


class TestInterface(unittest.TestCase):

    def setUp(self):
        self.nl = 'Test the nl_function with value set to 1.'
        self.trigger = pynlai.Trigger(
            core.to_obj,
            views._DEP_TOKEN['HR'],
            OrderedDict([
                ('lemma_', 'nl_function'),
                ('dep_', 'dobj'),
                ('head.lemma_', 'test'),
                ('head.pos_', 'VERB'),
            ]),
        )
        def arg_callback(sent):
            ents = core.to_ent(doc=sent, nlp=nlp).pop()
            view = core.create_view(ents, views._ENT_SPAN['HR'])
            return dict([('value', view['text'])])
        self.argument = pynlai.Argument(
            core.to_nc,
            views._DEP_SPAN['HR'],
            OrderedDict([
                ('root.lemma_', 'value'),
            ]),
            arg_callback,
        )
        @pynlai.nl_function(
            self.trigger,
            self.argument,
        )
        def nl_function(value):
            return value
        self.nl_function = nl_function

    def tearDown(self):
        pass

    def test_decorator(self):
        self.assertTrue(hasattr(self.nl_function, '__pynlai_triggers'))

    def test_run(self):
        r = pynlai.run(doc=self.nl, nlp=nlp, obj=self)
        self.assertEqual(r, u'1')



# -*- coding: utf-8 -*-

'''
core
----

core functions and classes for pynlai.
'''


from collections import OrderedDict
from functools import wraps
from operator import attrgetter
import six

from spacy.en import English
from spacy.tokens import Doc
from spacy.symbols import dobj, nsubj, VERB


def create_view(obj, view):
    '''
    filters object through view fields and returns field: value dict
    '''

    return OrderedDict([(f, attrgetter(f)(obj)) for f in view])


def nlp_preprocess(nlp_model):
    '''
    a decorator to pre-process text passed to nlp functions
    requires keyword args that at least contain doc and nlp
    '''

    def decorator(fcn):

        @wraps(fcn)
        def wrapper(*args, **kwargs):

            name = fcn.__name__

            kwargs.setdefault('doc', None)
            kwargs.setdefault('nlp', None)

            # validations
            if args:
                raise ValueError('pass args by name to %s' % name)

            if type(kwargs['doc']) not in (Doc, str, six.text_type):
                m = 'must pass `doc` arg of Doc or text type not %s'
                raise ValueError(m % type(kwargs['doc']))

            if type(kwargs['nlp']) is not nlp_model:
                m = 'must pass `nlp` arg of type %s'
                raise ValueError(m % nlp_model)

            # pre-processors
            if type(kwargs['doc']) is Doc:
                pass

            elif type(kwargs['doc']) is six.text_type:
                kwargs['doc'] = kwargs['nlp'](kwargs['doc'])

            else:
                kwargs['doc'] = kwargs['nlp'](six.u(kwargs['doc']))

            # call decorated function using pre-processed kwargs
            return fcn(**kwargs)

        return wrapper

    return decorator


@nlp_preprocess(English)
def find_dep(doc, dep, pos, **kwargs):
    '''
    finds dependencies in the sentence matching dep and (head) pos
    returns list of matching tokens
    '''

    return [t for t in doc if t.dep == dep and t.head.pos == pos]


@nlp_preprocess(English)
def to_ent(doc, **kwargs):
    '''
    returns entities and annotations
    see https://spacy.io/docs/usage/entity-recognition
    '''

    return [ent for ent in doc.ents]


@nlp_preprocess(English)
def to_nc(doc, **kwargs):
    '''
    returns noun chunks using spacy syntactic dependency parser
    see https://spacy.io/docs/usage/dependency-parse
    '''

    return [nc for nc in doc.noun_chunks]


@nlp_preprocess(English)
def to_obj(doc, **kwargs):
    '''
    finds the objects(s) of the sentence, which is the noun that is
    the recipient of a verb action
    '''

    return find_dep(doc=doc, dep=dobj, pos=VERB, **kwargs)


@nlp_preprocess(English)
def to_pos(doc, **kwargs):
    '''
    returns parts of speech from spacy POS tagger
    see https://spacy.io/docs/usage/pos-tagging
    '''

    return [word_token for word_token in doc]


@nlp_preprocess(English)
def to_sub(doc, **kwargs):
    '''
    finds the subject(s) of the sentence, which is the noun that is
    performing the verb(s) in the sentence
    '''

    return find_dep(doc=doc, dep=nsubj, pos=VERB, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
test_core
---------

unit tests for pynlai package core functionality
'''


import six
import unittest

from click.testing import CliRunner
import en_core_web_sm as en
from spacy.tokens import Doc

from pynlai import cli
from pynlai import core
from pynlai import views

from .shared import *


class TestCore(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    def tearDown(self):
        pass

    def test_nlp_preprocess(self):
        try:
            core.to_pos('Test.', nlp)
        except ValueError as e:
            self.assertTrue('pass args by name' in str(e))
        try:
            core.to_pos(nlp=nlp)
        except ValueError as e:
            self.assertTrue('must pass `doc`' in str(e))
        try:
            core.to_pos(doc='Test.')
        except ValueError as e:
            self.assertTrue('must pass `nlp`' in str(e))
        try:
            core.to_pos(doc=object(), nlp=nlp)
        except ValueError as e:
            self.assertTrue('arg of Doc or text type' in str(e))
        r = core.to_pos(doc=Doc(nlp.vocab), nlp=nlp)
        six.assertCountEqual(self, r, [])

    def test_create_view(self):
        class View(object):
            a = 1
            b = 2
            c = 3
        o = View()
        v = ('a', 'c')
        r = core.create_view(o, v)
        e = {'a': 1, 'c': 3}
        self.assertDictEqual(r, e)

    def test_to_ent(self):
        s = 'London is a big city in the United Kingdom.'
        v = views._ENT_SPAN['HR']
        r = [core.create_view(e, v) for e in core.to_ent(doc=s, nlp=nlp)]
        ents = [
            {'text': 'London', 'label_': 'GPE'},
            {'text': 'the United Kingdom', 'label_': 'GPE'},
        ]
        for i in range(0, 2):
            self.assertDictEqual(r[i], ents[i])

    def test_to_nc(self):
        s = 'I like green eggs and ham.'
        v = views._DEP_SPAN['HR']
        r = [core.create_view(e, v) for e in core.to_nc(doc=s, nlp=nlp)]
        ncs = [
            {
                'text': 'I',
                'root.orth_': 'I',
                'root.lemma_': '-PRON-',
                'root.dep_': 'nsubj',
                'root.head.orth_': 'like',
                'root.head.lemma_': 'like',
                'root.head.pos_': 'VERB',
            },
            {
                'text': 'green eggs',
                'root.orth_': 'eggs',
                'root.lemma_': 'egg',
                'root.dep_': 'dobj',
                'root.head.orth_': 'like',
                'root.head.lemma_': 'like',
                'root.head.pos_': 'VERB',
            },
            {
                'text': 'ham',
                'root.orth_': 'ham',
                'root.lemma_': 'ham',
                'root.dep_': 'conj',
                'root.head.orth_': 'eggs',
                'root.head.lemma_': 'egg',
                'root.head.pos_': 'NOUN',
            },
        ]
        six.assertCountEqual(self, r, ncs)

    def test_to_obj(self):
        s = 'I like green eggs and ham,'
        v = views._DEP_TOKEN['HR']
        r = [core.create_view(e, v) for e in core.to_obj(doc=s, nlp=nlp)]
        objs = [
            {
                'orth_': 'eggs',
                'lemma_': 'egg',
                'dep_': 'dobj',
                'head.orth_': 'like',
                'head.lemma_': 'like',
                'head.pos_': 'VERB',
            },
        ]
        six.assertCountEqual(self, r, objs)

    def test_to_pos(self):
        s = 'This is a test sentence.'
        v = views._POS_TOKEN['HR']
        r = [core.create_view(e, v) for e in core.to_pos(doc=s, nlp=nlp)]
        pos = [
            {'orth_': 'This', 'pos_': 'DET', 'lemma_': 'this'},
            {'orth_': 'is', 'pos_': 'VERB', 'lemma_': 'be'},
            {'orth_': 'a', 'pos_': 'DET', 'lemma_': 'a'},
            {'orth_': 'test', 'pos_': 'NOUN', 'lemma_': 'test'},
            {'orth_': 'sentence', 'pos_': 'NOUN', 'lemma_': 'sentence'},
            {'orth_': '.', 'pos_': 'PUNCT', 'lemma_': '.'},
        ]
        six.assertCountEqual(self, r, pos)

    def test_to_sub(self):
        s = 'I like green eggs and ham,'
        v = views._DEP_TOKEN['HR']
        r = [core.create_view(e, v) for e in core.to_sub(doc=s, nlp=nlp)]
        subs = [
            {
                'orth_': 'I',
                'lemma_': '-PRON-',
                'dep_': 'nsubj',
                'head.orth_': 'like',
                'head.lemma_': 'like',
                'head.pos_': 'VERB',
            },
        ]
        six.assertCountEqual(self, r, subs)



# -*- coding: utf-8 -*-
# flake8: noqa

'''
top-level package for pynlai.
'''


from .interface import (
    Trigger,
    Argument,
    nl_function,
    run,
)


__author__ = '''Chris Pappalardo'''
__email__ = 'cpappala@yahoo.com'
__version__ = '0.1.2'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
setup
-----

The pynlai setup script.
'''


from setuptools import setup, find_packages


with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY') as history_file:
    history = history_file.read()

requirements = [
    'click',
    'six',
    'spacy',
]

setup_requirements = [
]

test_requirements = [
    'mock',
]

DESCRIPTION = 'PYthon Natural Language Application Interface library.'

LONG_DESCRIPTION = '''
pynlai was created to provide a way for non-technical users to interact
with backend applications using natural language. Developers can simply
write an app and use pynlai to process text commands from any source
(e.g. irc, slack, email, etc.) by decorating their functions with natural
language triggers.

For more information and a guide for getting started, please see our
repo on github linked below.
'''

setup(
    name='pynlai',
    version='0.1.2',
    description="PYthon Natural Language Application Interface library.",
    long_description=LONG_DESCRIPTION + '\n\n' + history,
    author="Chris Pappalardo",
    author_email='cpappalardo@alvarezandmarsal.com',
    url='https://github.com/alvarezandmarsal/pynlai',
    packages=find_packages(include=['pynlai']),
    entry_points={
        'console_scripts': [
            'pynlai=pynlai.cli:entry_point'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='pynlai',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)
